#include "NMRsim.h"
#include "NMRsim_logfile.h"
#include "Parser.h"
#include "NMRsim_matrixspec.h"
#include "NMRsim_MasterObj.h"

matrixmap_type matrixmap;
prop_t prop_method=PROP_DIAG;
//bool have_spinsys=false;

#define MAXDT_TRIGGER 10
#define CHEBYSHEV_TRIGGER 5
#define ZEROTOLERANCE_TRIGGER 0.1

void make_par_variables();
static size_t lastrank=0;

int global_prop_flags=0;
double chebyshev_crossover=DEFAULT_CHEBYSHEV_CROSSOVER;
double eigtolerance=1e-6;
double zerotolerance=0.0; //don't round close-to-zero values by default

void parse_operator();

static const double rad_to_deg=180.0/M_PI;
static const double deg_to_rad=M_PI/180.0;

#ifndef DISABLE_EXEC
//#include "NMRsim_Process.h"
#include "NMRsim_spinsys.h"
#include "libcmatrix/Propagation.h"
#include "libcmatrix/AsynchronousPropagation.h"
#include "libcmatrix/timer.h"

void parse_sideband();
void parse_eigenvalue();
void parse_prop_method();
void parse_chebyshev_iterations();
void parse_zerotolerance();
void parse_maxdt();
void parse_steps_per_rotation();
void parse_maxjumpdt();
void parse_makeoperator();

acc_t acc_mode=ACC_TIME;
bool isfrequency() { return (acc_mode==ACC_FREQ); }

optional_t optsync;
optional_t optrealH;

namespace {
  struct Proxy_ {
    Proxy_() {
      optional_map_t& optional_map(get_optional_map());
      optional_map["minimalupdating"]=&optupdate;
      optional_map["synchronise"]=&optsync;
      optional_map["realhamiltonian"]=&optrealH;
      
      addprepar_callback(make_par_variables);
    }
  };  
  static Proxy_ proxy;
}

LIST<InternalCommand*> internal_timers;

InternalCommand* create_timer(const char* name)
{
  InternalCommand* timerp=new InternalCommand(name);
  internal_timers.push_back(timerp);
  return timerp;
}

InternalCommand* setup_cp=create_timer("Simulation set up");

BlockedOperator density; //!< current density matrix
int evaluation_index=0;
size_t quadrupole_order=1;
//size_t gamma_loop_steps=0;
double detect_freq=0.0;
double proton_freq=0.0;
double spin_rate=0.0;
int gamma_angles=0;
double def_rotor_angle=MAGIC_ANGLE; 
double rotor_orient=0.0;
CrystalStructure* cstructp=NULL;
double maxdtv=1.0e-6;
double maxjumpdt=0.0;
//bool forcegcompute=false;

int eigenvalue=NMRSIM_INVALID;
int eigenvalues=NMRSIM_INVALID;
int sideband=NMRSIM_INVALID;
double current_gamma=NMRSIM_INVALID;
spin_system* sysp=NULL;
Euler global_powder(0.0,0.0,0.0);
double gamma_zero=0.0;

double gammatimeoffset() {
  if ((spin_rate==0.0) || (current_gamma==NMRSIM_INVALID))
    throw InternalError("gammatimeoffset");
  return current_gamma/(TWO_PI*spin_rate);
}

//! NB look for exact zeros as these are likely to correspond to empty blocks
template<class M> void dump_density(const BlockedMatrix<M>& a, std::ostream& ostr =std::cout)
{
  if ((verbose & VER_PROFILE)==0)
    return;
  const BaseList<M> asrow(a.row());
  const size_t items=asrow.size();
  size_t nz=0;
  const M zero(0);
  for (size_t i=items;i--;) {
    if (asrow(i)!=zero)
      nz++;
  }
  ostr << "Density: " << (100.0*nz/items) << "%\n";
}

//! output blocked matrix ('spy' if precision 0)
template<class M> void spyprint(std::ostream& ostr,const BlockedMatrix<M>& a) {
  //  if (precision)
  ostr << a;
  //  else
  //spy(ostr,a);
  dump_density(a,ostr);
}

//! output operator matrix ('spy' if precision 0)
void spyprint(std::ostream& ostr,const BlockedOperator& a)
{
  //  if (precision)
    ostr << a;
//   else
//     spy(ostr,a.row());
  dump_density(a.row(),ostr);
}

//! default - 0 precision not special 
template<class M> void spyprint(std::ostream& ostr,const M& a) {
  ostr << a;
}

MasterObj* masterobjp=NULL;
bool istimedomain;

double get_rotor_phase(double t)
{
  if (!masterobjp)
    throw InternalError("get_rotor_phase: called before Hamiltonian created");
  static const double lim360=1.0-1e-8;
  
  try {
    const double x=masterobjp->phase(t)/TWO_PI;
    const double frac=x-floor(x);
    return (frac>lim360) ? 0.0 : 360.0*frac;
  } catch (MatrixException&) {
    return 0.0;
  }
}

inline double get_rotor_phase() { return get_rotor_phase(master_time); }

static timer<> global_stopwatch;
inline double get_cputime() { return global_stopwatch(); }

// struct SysVar_chebyshev : public SystemVariable<size_t*> {
//   SysVar_chebyshev()
//     : SystemVariable<size_t*>("chebyshev_iterations",&(cmatrix_eigensystem_controller.chebyshev_iterations),false,true) {}
  
//   void update();
// };
  
// void SysVar_chebyshev::update() 
// {
//   if ((*this)()>eigensystem_controller::max_chebyshev_iterations) {
//     std::cerr << "Maximum Chebyshev iterations is " << eigensystem_controller::max_chebyshev_iterations <<'\n';
//     error_abort();
//   }
// }

SystemVariable<double*> v_spin_rate("spin_rate",&spin_rate,1.0,V_UPDATEINTS | V_ISCONST); //change to spin rate forces Hamiltonian re-build
SystemVariable<double*> v_rotor_angle("rotor_angle",&def_rotor_angle,rad_to_deg,V_UPDATEINTS | V_ISCONST);
SystemVariable<int*> v_gamma_angles("gamma_angles",&gamma_angles,V_ISCONST);
SystemVariable<double*> v_gammazero("gamma_zero",&gamma_zero,rad_to_deg,V_UPDATEINTS | V_ISCONST);
SystemVariable<double*> v_proton_freq("proton_frequency",&proton_freq,1.0,V_ISCONST | V_UPDATEINTS);
SystemVariable<double*> v_observe_freq("observe_frequency",&detect_freq,1.0,V_ISFIXED);
//SystemVariable<int*> v_sum_index("i_sum",&sum_index,true,false); //fixed, but not constant
SystemVariable<int*> v_evaluation_index("i_evaluation",&evaluation_index,V_ISFIXED);
SystemVariable<int*> v_sideband("sideband",&sideband,0);
SystemVariable<int*> v_eigenvalue("eigenvalue",&eigenvalue,0);
SystemVariable<int*> v_eigenvalues("eigenvalues",&eigenvalues,V_ISFIXED | V_ISCONST);
SystemVariable<size_t*> v_chebyshev("chebyshev_iterations",&(cmatrix_eigensystem_controller.chebyshev_iterations),V_ISCONST);
//SysVar_chebyshev v_chebyshev;
//SystemVariable<int*> v_ntimes("acqseq_ntimes",&acqseq_ntimes,false,true);
SystemVariable<double*> v_time("time",&master_time,1e6,V_ISFIXED); //output time in us
//SysVar_trans v_trans;
SystemVariable<FUNC_REAL> v_rotor_phase("rotor_phase",get_rotor_phase);
SystemVariable<FUNC_REAL> v_cputime("cputime",get_cputime);

bool isallowed_nonconst_var(SystemVariableBase* varp)
{
  return (varp==&v_evaluation_index);
}

bool proton_freq_isconstant()
{
  return v_proton_freq.isconstant();
}

BlockedOperator sigma0,detect;
LIST<nuclei_spec> blockingnuclei;

Warning<> syncfailed_warning("failed to find synchronisation",&NMRsim_repeat_warning);

void make_par_variables()
{
  if (optsync==OPTIONAL_OFF)
    syncfailed_warning.type(BaseWarning::Ignore); //!< disable warnings if synchronisation explicitly disabled
  
  make_1d_par_variables();

  add_systemvarmap(v_spin_rate);
  add_systemvarmap(v_rotor_angle);
  add_systemvarmap(v_gamma);
  add_systemvarmap(v_gamma_angles);
  add_systemvarmap(v_observe_freq);
  add_systemvarmap(v_eigenvalue);
  add_systemvarmap(v_eigenvalues);
  add_systemvarmap(v_sideband);  
  //  add_systemvarmap(v_sum_index);
  add_systemvarmap(v_evaluation_index);

  command_Factory_t& par_Factory(get_par_Factory());
  par_Factory["maxdt"]=par_t(&parse_maxdt,true);
  par_Factory["maxjumpdt"]=par_t(&parse_maxjumpdt,true);
  par_Factory["steps_per_rotation"]=&parse_steps_per_rotation;
  //  par_Factory["operator"]=par_t(&parse_makeoperator,true);
  par_Factory["eigenvalue"]=&parse_eigenvalue;
  par_Factory["sideband"]=&parse_sideband;  
  par_Factory["propagation_method"]=&parse_prop_method;
  par_Factory["chebyshev_iterations"]=&parse_chebyshev_iterations;
  par_Factory["zerotolerance"]=&parse_zerotolerance;
}

template<class T> inline bool hasnonzero(const T& a) {
  const typename T::const_iterator aend=a.end();
  const typename T::value_type zero(0);
  return (std::find_if(a.begin(),aend,std::bind1st(std::not_equal_to<typename T::value_type>(),zero))!=aend);
}

Warning<> matrixspec::writefailed_warning("write to log file failed: object not defined at this point",&NMRsim_repeat_warning);

void matrixspec::log(logfile_controller& ctrl, const char* name, bool full) const
{
  try {
    switch (type) {
    case OPERATOR: {
	const BlockedOperator& op(asoperator().op); 
	if (full) {
	  cmatrix tmp;
	  assert(masterobjp!=NULL);	  
	  masterobjp->expandoperator(tmp,op);
	  ctrl.write(tmp,name);
	}
	else
	  ctrl.write(op.row(),name);
    }
      break;
//     case DOUBLE:
//       ctrl.write(asdouble(),name);
//       break;
    case BOOL:
      ctrl.write(asfilter()().row(),name);
      break;
    default:
      throw InternalError("matrixspec: can only log operator or filter matrices");
    }
  } catch (Undefined&) {
    writefailed_warning.raise();
  }
}

std::ostream& operator<< (std::ostream& ostr, const matrixspec& curm)
{
  switch (curm.type) {
  case matrixspec::SPECIAL:
    return ostr << "Can't output here (local)\n";
  case matrixspec::BOOL: {
    const BlockedFilter* filterp(curm.asfilter().filterp);
    if (filterp)
      return ostr << *filterp;
    else
      return ostr << "<undefined here>\n";
  }
  case matrixspec::OPERATOR:
    spyprint(ostr,curm.asoperator().op);
    return ostr;
//   case matrixspec::DOUBLE:
//     spyprint(ostr,curm.asdouble());
//     return ostr;
  case matrixspec::EXCHANGE:
    return ostr << curm.asexchange();
  case matrixspec::NONE:
    return ostr << "<undefined>\n";
  }
  throw InternalError("Unknown matrix type");
}

usage_t matrixspec::usage() const
{
  switch (type) {
  case BOOL:
    return usage_(asfilter()());
  case OPERATOR:
    return usage_(asoperator().op.row());
    //  case DOUBLE:
    //return usage_(asdouble());
  case EXCHANGE:
    return usage_(asexchange()());
  }
  return usage_t(); //!< 0 for special types    
}

matrixspec& matrixspec::operator= (const matrixspec& from)
{
  switch (from.type) {
  case BOOL:
    set_filter(from.asfilter());
    return *this;
  case OPERATOR:
    set_operator(from.asoperator());
    return *this;
//   case DOUBLE:
//     set_double(from.asdouble());
//     return *this;
  case NONE:
    type=NONE;
    ptr=NULL;
    return *this;
  case EXCHANGE:
    set_exchange(from.asexchange());
    return *this;
  case SPECIAL:
    type=SPECIAL;
    ptr=from.ptr;
    return *this;
  default:
    throw Undefined("matrixspec=");
  }
}      

bool matrixspec::operator!() const 
{
  switch (type) {
  case OPERATOR:
    return !(asoperator().op);
  case BOOL:
    return !asfilter();
  case EXCHANGE:
    return !(asexchange()());
    //  case DOUBLE:
    // return static_cast< const BlockedMatrix<double>* >(ptr)->empty();
  case SPECIAL:
    return false;
  }
  return true;
}

void matrixspec::set_operator(const operator_def& a)
{
  type=OPERATOR;
  ptr=static_cast<const void*>(&a);
}

// void matrixspec::set_double(const BlockedMatrix<double>& a)
// {
//   type=DOUBLE;
//   ptr=static_cast<const void*>(&a);
// }

void matrixspec::set_exchange(const ExchangeMatrix& a)
{
  type=EXCHANGE;
  ptr=static_cast<const void*>(&a);
}

void matrixspec::set_filter(const filter_def& a)
{
  type=BOOL;
  ptr=static_cast<const void*>(&a);
}

const operator_def& matrixspec::asoperator() const
{
  if (type!=OPERATOR)
    throw Failed("matrixspec::asoperator: not an operator matrix");
  return *static_cast< const operator_def* >(ptr);
}

const ExchangeMatrix& matrixspec::asexchange() const
{
  if (type!=EXCHANGE)
    throw Failed("matrixspec::asexchange: not an exchange matrix");
  return *static_cast< const ExchangeMatrix* >(ptr);
}

// const BlockedMatrix<double>& matrixspec::asdouble() const 
// {
//   if (type!=DOUBLE)
//     throw Failed("matrixspec::ascomplex: not an rmatrix");
//   return *static_cast< const BlockedMatrix<double>* >(ptr);
// }

const filter_def& matrixspec::asfilter() const
{
  if (type!=BOOL)
    throw Failed("matrixspec::ascomplex: not a filter matrix");
  return *static_cast<const filter_def*>(ptr);
}

// template<class R,class C> matrixspec deref(const RealComplexHolder<R,C>& a)
// {
//   typedef RealComplexHolder<R,C> holder_t;
//   switch (a.type()) {
//   case holder_t::REAL:
//     return matrixspec((*(a.get_real()))());
//   case holder_t::COMPLEX:
//     return matrixspec((*(a.get_complex()))());
//   }
//   throw Undefined("dereference Hamiltonian");
// }

template<class R,class C> bool isempty(const RealComplexHolder<R,C>& a)
{
  typedef RealComplexHolder<R,C> holder_t;
  switch (a.type()) {
  case holder_t::REAL:
    return !(*(a.get_real()));
  case holder_t::COMPLEX:
    return !(*(a.get_complex()));
  }
  return true;
}

bool MasterObj::Hempty() const
{
  switch (hamobjtype) {
  case M_STATIC:
    return isempty(Hstaticp);
  case M_STATICD:
    return !(*Hstaticdp);
  case M_SPIND:
    return !(*Hspindp);
  case M_SPIN:
    return isempty(Hspinp);
  default:
    throw InternalError("Hempty");
  }
}

usage_t MasterObj::usage() const
{
  return Husage();
}

usage_t MasterObj::Husage() const
{
  switch (hamobjtype) {
  case M_STATIC:
    switch (Hstaticp.type()) {
    case Hstatic_t::REAL:
      return usage_t((*(Hstaticp.get_real()))().row().size(),Type2Type<double>());
    case Hstatic_t::COMPLEX:
      return usage_t((*(Hstaticp.get_complex()))().row().size(),Type2Type<complex>());
    }
    break;
  case M_STATICD:
    return usage_t(Hstaticdp->usage(),Type2Type<double>());
  case M_SPIND:
    return usage_t(Hspindp->usage(),Type2Type<double>());
  case M_SPIN:
    switch (Hspinp.type()) {
    case Hspin_t::REAL:
      return usage_t((Hspinp.get_real())->usage(),Type2Type<double>());
    case Hspin_t::COMPLEX:
      return usage_t((Hspinp.get_complex())->usage(),Type2Type<complex>());
    }
  }
  throw InternalError("Unknown/unset hamobjtype");
}


void MasterObj::printH(std::ostream& ostr) const
{
  switch (hamobjtype) {
  case M_STATIC:
    switch (Hstaticp.type()) {
    case Hstatic_t::REAL:
      spyprint(ostr,*(Hstaticp.get_real()));
      break;
    case Hstatic_t::COMPLEX:
      spyprint(ostr,*(Hstaticp.get_complex()));
      break;
    }
    break;
  case M_STATICD:
    ostr << (*Hstaticdp) << '\n';
    break;
  case M_SPIND:
    spyprint(ostr,(*Hspindp)(master_time));
    break;
  case M_SPIN:
    switch (Hspinp.type()) {
    case Hspin_t::REAL:
      spyprint(ostr,(*(Hspinp.get_real()))(master_time));
      break;
    case Hspin_t::COMPLEX:
      spyprint(ostr,(*(Hspinp.get_complex()))(master_time));
      break;
    default:
      ostr << "<unset>\n";
    }
    break;
  default:
    throw Failed("Unknown hamobjtype");
  }
  if (verbose & VER_PROFILE)
    ostr << Husage();
}

void MasterObj::printfullH(std::ostream& ostr) const
{
  if (!spin_rate) {
    printH(ostr);
    return;
  }
  switch (hamobjtype) {
  case M_SPIND:
    ostr << *Hspindp;
    break;
  case M_SPIN:
    switch (Hspinp.type()) {
    case Hspin_t::REAL:
      ostr << *(Hspinp.get_real());
      break;
    case Hspin_t::COMPLEX:
      ostr << *(Hspinp.get_complex());
      break;
    default:
      ostr << "<unset>\n";
    }
    break;
  default:
    throw Failed("Unknown hamobjtype");
  }
  if (verbose & VER_PROFILE)
    ostr << Husage();
}

namespace {
  void write_diag(logfile_controller& ctrl, const ListList<double>& Hdiag, const char* name)
  {
    BlockedMatrix<double> Hfull;
    full(Hfull,Hdiag);
    ctrl.write(Hfull,name);
  }
}

void MasterObj::logH(logfile_controller& ctrl, const char* name) const
{
  switch (hamobjtype) {
  case M_STATIC:
    switch (Hstaticp.type()) {
    case Hstatic_t::REAL:
      ctrl.write((*(Hstaticp.get_real()))(),name);
      break;
    case Hstatic_t::COMPLEX:
      ctrl.write((*(Hstaticp.get_complex()))(),name);
      break;
    }
    //    deref(Hstaticp).log(ctrl,name);
    return;
  case M_STATICD: 
    write_diag(ctrl,(*Hstaticdp)(),name);
    return;
  case M_SPIND: {
    const ListList<double> Hdiag((*Hspindp)(master_time));
    write_diag(ctrl,Hdiag,name);
  }
    return;
  case M_SPIN:
    switch (Hspinp.type()) {
    case Hspin_t::REAL:
      ctrl.write((*(Hspinp.get_real()))(master_time),name);
      return;
    case Hspin_t::COMPLEX:
      ctrl.write((*(Hspinp.get_complex()))(master_time),name);
      return;
    }
  default:
    throw Failed("Unknown/undefined Hamiltonian");
  }
}

template<class OpGen> void MasterObj::create_Hstatic(const OpGen& opgen, Type2Type<complex>)
{
  (Hstaticp.set_complex()).reset(new BlockedStaticHamiltonian<complex>(opgen));
}

template<class OpGen> void MasterObj::create_Hstatic(const OpGen& opgen, Type2Type<double>)
{
  (Hstaticp.set_real()).reset(new BlockedStaticHamiltonian<double>(opgen));
}

template<class OpGen> void MasterObj::create_Hspin(const OpGen& opgen, double gamma,Type2Type<complex>)
{
  (Hspinp.set_complex()).reset(new BlockedSpinningHamiltonian<complex>(opgen,spin_rate,gamma,rinfo));//,CWdesc));
}

template<class OpGen> void MasterObj::create_Hspin(const OpGen& opgen, double gamma,Type2Type<double>)
{
  (Hspinp.set_real()).reset(new BlockedSpinningHamiltonian<double>(opgen,spin_rate,gamma,rinfo));
}

bool MasterObj::reset_angle(double angle)
{
  if (spin_rate) {
    const size_t rank= Hstructp->isclassicsecondorder() ? 4 : 2;
    if ((rank!=lastrank) || (angle!=rotor_angle)) {
      if ((verbose & VER_GEN) && (verbose_level>1))
	std::cout << "Changing/setting rotor angle to " << (angle*rad_to_deg) << " degrees (rank " << rank << ")\n";
      rinfo=RotorInfo(rank,angle,rotor_orient);
      rotor_angle=angle;
      lastrank=rank;
      return true;
    }
  }
  else
    rotor_angle=0.0;
  return false;
}
 
bool MasterObj::change_angle(double angle)
{
  const bool changed_angle=reset_angle(angle);
  if (changed_angle)
    recreate_Hs();
  return changed_angle;
}

template<class HType,class StoreType> void set_interactions(HType& H, const Euler& powder, Type2Type<StoreType>)
{
  HamiltonianStore<StoreType> Hstore(*interactions_MFp,powder);
  if (verbose & VER_GEN)
    std::cout << Hstore;
  H.interactions(Hstore);
}    

template<class StaticObj,class SpinObj> void MasterObj::set_interactions(const Euler& powder, StaticObj& Hstaticp, SpinObj& Hspinp)
{
  switch (hamobjtype) {
  case M_SPIND:
    // Need to trap empty Hamiltonian otherwise propagation will fail
    if (interactions_MFp->empty())
      error_abort("Inhomogeneous Hamiltonian but no interactions - what is there to simulate?");
    ::set_interactions(*Hspindp,powder,Type2Type<space_T>());
    break;
  case M_SPIN:
    ::set_interactions(*Hspinp,powder,Type2Type<space_T>());
    break;
  case M_STATICD:
    if (quadrupole_order!=1)
      ::set_interactions(*Hstaticdp,powder,Type2Type<space_T>());
    else 
      ::set_interactions(*Hstaticdp,powder,Type2Type<double>());    
    break;
  case M_STATIC:
    if (quadrupole_order!=1)
      ::set_interactions(*Hstaticp,powder,Type2Type<space_T>());
    else {
      ::set_interactions(*Hstaticp,powder,Type2Type<double>());
      set_Hsystem((*Hstaticp)());
    }
    break;
  default:
    throw Failed("Unknown hamobjtype");
  }

  if (verbose & VER_GEN) {
    std::cout << "Full Hamiltonian\n";
    printfullH(std::cout);
  }
}

/**
   \pre Hamiltonian must be static (otherwise \c Failed exception)
*/
void MasterObj::rotor_phase(double phase0)
{
  switch (hamobjtype) {
  case M_SPIN:
    if (Hspinp.iscomplex())
      (Hspinp.get_complex())->rotor_phase(phase0);
    else
      (Hspinp.get_real())->rotor_phase(phase0);
    break;
  case M_SPIND:
    Hspindp->rotor_phase(phase0);
    break;
  default:
    throw Failed("rotor_phase called for static Hamiltonian!");
  }
}

/**
   \pre Hamiltonian must be static (otherwise \c Failed exception)
*/
double MasterObj::phase(double t) const
{
  switch (hamobjtype) {
  case M_SPIN:
    return (Hspinp.iscomplex()) ?
      (Hspinp.get_complex())->phase(t) : (Hspinp.get_real())->phase(t);
  case M_SPIND:
    return Hspindp->phase(t);
  }
  throw Failed("phase called for static Hamiltonian!");
}

int check_sync(double nfloat,double tol)
{
  if (nfloat<=0.0) {
    char buf[256];
    snprintf(buf,sizeof(buf),"check_sync: asked to round to an integer a number (%g) which is <=0",nfloat);
    throw InvalidParameter(buf);
  }
  if (optsync==OPTIONAL_OFF)
    return 0;
  int n=int(nfloat+0.5);
  return (fabs(nfloat-n)>tol) ? 0 : n;
}

bool MasterObj::canbuildoperators(const BaseList<productoperator_spec*>& specs) const
{
  for (size_t i=specs.size();i--;) {
    const productoperator_spec& curspec(*(specs(i)));
    try {
#ifndef NOPERIODIC
      if (!simple_opgenp)
	block_pattern bp(*crystal_opgenp,curspec);
      else
#endif
	block_pattern bp(*simple_opgenp,curspec);

    } catch (Failed&) {
      if (verbose & VER_GEN)
	std::cerr << "Failed to build spin operator: " << curspec << '\n';
      return false;
    }
  }
  return true;
}

MasterObj::MasterObj()
{
}

Warning<> commonPAS_warning("Optimisation for special case of common PAS is not implemented",&NMRsim_once_warning);

MasterObj::MasterObj(const basespin_system& sys,const HamiltonianStore<space_T>& Hstorev,const BaseList<nuclei_spec>& blockingnuc,int flags, bool use_crystal)
    : 
  Hstructp(new HamiltonianStructure(sys,Hstorev,blockingnuc,weakints,flags)), 
  simple_opgenp(use_crystal ? NULL : new SpinOpGenerator(*Hstructp,*cstructp,(verbose & VER_GEN) ? verbose_level : 0))
#ifndef NOPERIODIC
  ,crystal_opgenp(use_crystal ? new CrystalOpGenerator(*Hstructp,*cstructp,(verbose & VER_GEN) ? verbose_level : 0) : NULL)
#endif
  ,hamtype(H_NULL), hamobjtype(M_NULL)
{
  if (!canbuildoperators(opspecstack))
    throw Failed("Can't build spin operators");

  update();

  if (!have_spinsys()) 
    return; 
  //    restart();
  //  return;
  //}
  
  if (verbose & VER_GEN) {
    std::cout << "Hamiltonian structure:\n" << *Hstructp;
#ifndef NOPERIODIC
    if (!simple_opgenp)
      std::cout << (*crystal_opgenp);
    else
#endif
      std::cout << (*simple_opgenp);
  }

//   if (Hstructp->isdiagonal())
//     ishomogeneous=false;
  
//   if (verbose & VER_GEN) {
//     std::cout << "Inhomogeneous: "; 
//     if (ishomogeneous)
//       std::cout << "No\n";
//       else {
// 	if (iscommonPAS)
// 	  std::cout << "Yes (common PAS)\n";
// 	else
// 	  std::cout << "Yes\n";
//       }
//   }
  
  if (sw && (verbose & VER_GEN)) {
    std::cout << "Acquisition time: ";
    prettyprint_time(np/sw) << '\n';
  }
  
  quadrupole_order=Hstructp->quadrupole_order();

  ensure_channels();
  //  restart();
}

//! these warnings output on cout rather than cerr
Warning<> generalisedqpole_warning("The generalised quadrupole treatment has not been fully evaluated.  Use at your own risk!",&NMRsim_once_warning,BaseWarning::Inherit,std::cout);
Warning<> generalisedqpole_rf_warning("RF sequences have been defined with non-first-order quadrupoles present.  Treatment is only approximate and has not been fully evaluated.  Use at your own risk!",&NMRsim_once_warning,BaseWarning::Inherit,std::cout);

void MasterObj::update()
{
  if (!have_spinsys())
    return;

  //!< can't do this in restart since havestore is not set until par block (after MasterObj created)
  if (quadrupole_order!=1) {
    if (havestore)
      generalisedqpole_rf_warning.raise();
    if (!isclassicQ())
      generalisedqpole_warning.raise();
  }

  if (proton_freq && !!simple_opgenp)
    simple_opgenp->proton_frequency(proton_freq);
  eigenvalues=spinop_generator().eigblocks();
}

BlockedOperator MasterObj::make(const productoperator_spec& spec) const
{
#ifndef NOPERIODIC
  if (!simple_opgenp)
    return BlockedOperator(*crystal_opgenp,spec);
#endif
  return BlockedOperator(*simple_opgenp,spec);
}


void MasterObj::restart()
{
  timer_guard_t guard(*setup_cp,true); //!< ignore if already timing

// //   if (phasep)
// //     phasep->update();
  update();

// //   if ((verbose & VER_GEN) && acqseqp)
// //     std::cout << "Acquisition sequence\n" << (*acqseqp) << '\n';

  if (have_spinsys())
    reset_angle(def_rotor_angle);
//   //  if (acqp)
//   //  acqp->checksync();
}

#ifndef NOPERIODIC
void MasterObj::recreate_Hs(const CrystalOpGenerator& opgen, double gamma)
{
  switch (hamobjtype) {
  case M_STATIC:
    create_Hstatic(opgen,Type2Type<complex>());
    break;
  case M_SPIN:
    create_Hspin(opgen,gamma,Type2Type<complex>());
    break;
  default:
    throw Failed("Unknown/unsupported hamobjtype");
  }
}
#endif

void MasterObj::recreate_Hs(const SpinOpGenerator& opgen, double gamma)
{
  switch (hamobjtype) {
  case M_STATIC:
    if (hamtype==H_COMPLEX)
      create_Hstatic(opgen,Type2Type<complex>());
    else
      create_Hstatic(opgen,Type2Type<double>());
    break;
  case M_SPIN:
    if (hamtype==H_COMPLEX)
      create_Hspin(opgen,gamma,Type2Type<complex>());
    else
      create_Hspin(opgen,gamma,Type2Type<double>());
    break;
  case M_SPIND:
    Hspindp.reset(new BlockedDiagonalSpinningHamiltonian(opgen,spin_rate,gamma,rinfo));
    break;
  case M_STATICD:
    Hstaticdp.reset(new BlockedDiagonalStaticHamiltonian(opgen));
    break;
  default:
    throw Failed("Unknown/unset hamobjtype");
  }
}

/**
   \post ::MasterObj::Hempty is \c false
*/
void MasterObj::recreate_Hs()
{
  static hamobj_t lasthamobjtype=M_NULL;
  static ham_t lasthamtype=H_NULL;

  if ((verbose & VER_GEN) && (verbose_level>1))
    std::cout << "Recreating Hamiltonian\n";

  //const CompSequenceBase* acqseqp = acqp ? acqp->sequence() : NULL;
  hamtype=H_COMPLEX;
  hamobjtype=M_NULL;

  //  if (!acqseqp && !noshortcuts && !haveactions) { //can't optimise if RF present
  const bool allowreal=!(Hstructp->iscomplex()) && optrealH;
  check_optional(optrealH,"real Hamiltonian",allowreal,"realhamiltonian");
  if (allowreal) {
    //    const bool allowdiagonal=!acqseqp && !haveactions && !!(phasep->simple_opgenp); //CrystalOpGen doesn't support diagonal 
    //    if (Hstruct.isdiagonal() && !hasacquisition_rf() && !haveactions && (quadrupole_order!=0))
    if (Hstructp->isdiagonal() && !haveactions && (quadrupole_order!=0))
      hamtype=H_DIAGONAL;
    else {
      if (quadrupole_order==1)
	hamtype=H_REAL;
    }
  }

  if (hamtype!=lasthamtype) {
    if (verbose & VER_GEN) {
      std::cout << "Hamiltonian type: ";
      switch (hamtype) {
      case H_DIAGONAL:
	std::cout << "diagonal\n";
      break;
      case H_COMPLEX:
	std::cout << "complex\n";
	break;
      case H_REAL:
	std::cout << "real\n";
	break;
      }
    }
    lasthamtype=hamtype;
  }

  if (spin_rate) {
    if (lastrank==0)
      throw InternalError("recreate_Hs called before RotorInfo established");

    switch (hamtype) {
    case H_DIAGONAL:
      hamobjtype=M_SPIND;
      break;
    default:
      if (iscommonPAS && !haveactions)
	commonPAS_warning.raise();
      hamobjtype=M_SPIN;
      break;
    }
  }
  else {
    switch (hamtype) {
    case H_DIAGONAL:
      hamobjtype=M_STATICD;
      break;
    case H_COMPLEX: case H_REAL:
      hamobjtype=M_STATIC;
      break;
    }
  }

#ifndef NOPERIODIC
  if (!simple_opgenp)
    recreate_Hs(*crystal_opgenp,global_powder.gamma);
  else
#endif
    recreate_Hs(*simple_opgenp,global_powder.gamma);
  
  const Euler powder_no_gamma(global_powder.alpha,global_powder.beta,0.0); //!< gamma angle is set in rotor phase
  if (iscomplex())
    set_interactions(powder_no_gamma,Hstaticp.get_complex(),Hspinp.get_complex());
  else
    set_interactions(powder_no_gamma,Hstaticp.get_real(),Hspinp.get_real());  

  if (lasthamobjtype!=hamobjtype) {
    verify_powderaverage(); //only verify powder average if changing type of Hamiltonian
    lasthamobjtype=hamobjtype;
  }
}

void parse_eigenvalue()
{
  parse_system_variable(v_eigenvalue);
}

void parse_sideband()
{
  parse_system_variable(v_sideband);
}

ContextWarning<> zerotolerance_warning("zerotolerance should be a small value",&NMRsim_once_warning);

void parse_zerotolerance()
{
  if (are_left()) {
    zerotolerance=parse_double();
    if (zerotolerance<0.0)
      error_abort("zerotolerance must be >=0.0");
    if (zerotolerance>ZEROTOLERANCE_TRIGGER)
      zerotolerance_warning.raise();
  }
  else
    std::cout << "zerotolerance=" << zerotolerance << '\n';
}
    
void parse_prop_method()
{
  const char* meth=parse_string();
//   if (strcmp(meth,"automatic")==0) {
//     prop_method=PROP_AUTO;
//     if (are_left()) {
//       chebyshev_crossover=0.01*parse_double();
//       if ((chebyshev_crossover<0.0) || (chebyshev_crossover>1.0))
// 	error_abort("chebyshev crossover density should be between 0 and 100 (%)\n");
//       if (allowwarnings() && (chebyshev_crossover!=0.0) && (chebyshev_crossover<0.01))
// 	std::cerr << "Warning: chebyshev crossover less than 1%  Value specified as fraction rather than percentage? (-nochecks disables this warning)\n";
//     }
//   }
//   else {
    if (strcmp(meth,"diagonalise")==0)
      prop_method=PROP_DIAG;
    else {
    if (strcmp(meth,"chebyshev")==0)
      prop_method=PROP_CHEBYSHEV;
    else
      error_abort("Unknown propagation method.  Known: diagonalise, chebyshev");
    }
    //  }
}

ContextWarning<> chebyshev_iterations_warning("chebyshev_iterations set to less than " NMRSIM_STRINGISE(CHEBYSHEV_TRIGGER) "; this is unlikely to work properly",&NMRsim_once_warning);

void parse_chebyshev_iterations()
{
  if (!are_left())
    std::cout << "chebyshev_iterations=" << v_chebyshev() << '\n';
  else {
    parse_system_variable(v_chebyshev);
    if (v_chebyshev()<CHEBYSHEV_TRIGGER)
      chebyshev_iterations_warning.raise();
  }
}

void parse_spin_rate()
{
  parse_system_variable(v_spin_rate,F_DENYZERO);
}

void parse_gamma_zero()
{
  parse_system_variable(v_gammazero);
}

tstep_t tstep_type=INT_UNSET;
int steps_per_rotation=0;

double get_maxdt()
{
  if (tstep_type==INT_ROTORPERIOD) {
    if (spin_rate==0)
      error_abort("integration timestep cannot be specified by steps_per_rotation in static simulation");
    return (1.0/spin_rate)/steps_per_rotation;
  }
  return maxdtv;
}

void set_inttype(tstep_t newtype)
{
  if (tstep_type && (tstep_type!=newtype))
    error_abort("cannot mix steps_per_rotation and maxdt when specifying integration time step");
  tstep_type=newtype;
}

ContextWarning<> maxdt_warning("maxdt set to >" NMRSIM_STRINGISE(MAXDT_TRIGGER) " us - are you sure about this?",&NMRsim_once_warning);
ContextWarning<> parameteroverwrite_warning("parameter (maxdt/maxjumpdt) set here overrides earlier setting",&NMRsim_repeat_warning); 

double parsedt(const char* parname, bool& doneset)
{
  if (doneset)
    parameteroverwrite_warning.raise();
  else
    doneset=true;
  const double dt=parse_double()*1e-6;
  if (dt<=0.0) {
    parser_printcontext() << parname << " must be >0\n";
    error_abort();
  }
  return dt;
}

void parse_maxdt()
{
  static bool doneset=false;
  if (!are_left())
    std::cout << "maxdt=" << (maxdtv*1e6) << " us\n";
  else {
    set_inttype(INT_EXPLICIT);
    maxdtv=parsedt("maxdt",doneset);
    if (maxdtv>MAXDT_TRIGGER)
      maxdt_warning.raise();
  }
}

void parse_steps_per_rotation()
{
  set_inttype(INT_ROTORPERIOD);
  steps_per_rotation=parse_int();
  if (steps_per_rotation<1)
    error_abort("steps_per_rotation must be >0");
}

void parse_maxjumpdt()
{
  static bool doneset=false;
  if (!are_left())
    std::cout << "maxjumpdt=" << (maxjumpdt*1e6) << " us\n";
  else
    maxjumpdt=parsedt("maxjumpdt",doneset);
}

void parse_gamma_angles()
{
  //  static flagsmap_type flags;
//   if (flags.empty())
//     flags["forcegcompute"]=1;
  parse_system_variable(v_gamma_angles);
  //  forcegcompute=(parse_flags(flags)!=0);
}

ContextWarning<> zeroangle_warning("rotor_angle of zero makes little sense!",&NMRsim_once_warning);

void parse_rotor_angle()
{
  //  if (are_left()) {
  parse_system_variable(v_rotor_angle);
  if (def_rotor_angle==0.0)
    zeroangle_warning.raise();
  //}
  //else
  //  std::cout << "rotor_angle=" << (def_rotor_angle*rad_to_deg) << " degrees\n";
}


void parse_generalmatrix(const char* name, bool isgeneral =true)
{
  //  const size_t nargs=count_left();
  //if (nargs<2)
  //  error_abort("Too few arguments to matrix set <name> exchange <real matrix elements> [<imaginary elements>]\n");
  //const size_t sites=(nargs>2) ? nargs-1 : 2;
  Mark markobj;
  Variable cvar(S_ARG1);
  VariableBase* elsp=parse_double_variable(cvar,F_ALLOWLIST);
  bool isconst=elsp->isconst();
  //  ExchangeMatrix* matrixp=NULL;
//   if (are_left()) {
//     if (!isgeneral)
//       error_abort("cannot specify imaginary component to exchange matrix (use matrix set general to create a general complex (square) matrix)");
//     cvar.subsid=S_ARG2;
//     VariableBase* imagp=parse_double_variable(cvar,F_ALLOWLIST);
//     if (!(imagp->isconst()))
//       isconst=false;
//     matrixp=new ExchangeMatrix(name,realp->value(),imagp->value(),isconst);
//   }
//   else
  ExchangeMatrix* matrixp=new ExchangeMatrix(name,elsp->value(),isconst,isgeneral);
  markobj.flush(matrixp);
  matrixmap[name]=matrixspec(*matrixp);
}

bool matchtag(const char* name)
{
  const char* ptr=get_curline();
  const size_t len=strlen(name);
  if (strncmp(ptr,name,len)!=0)
    return false;
  const char term=ptr[len];
  const bool found=(term=='\0') || isspace(term);
  if (found)
    (void)get_token(); //!< swallow token
  return found;
}
  
void parse_matrix()
{
  if (strcmp(parse_string(),"set")!=0)
    error_abort("syntax: matrix set <to> <from>");

  const char* matname= parse_string();

  if (strcmp(matname,"detect")==0)
    error_abort("use detect_operator to set detection operator");
  if (strcmp(matname,"start")==0)
    error_abort("use start_operator to set initial density matrix");

  if (matrixmap.count(matname))
    error_abort("can't redefine matrices");

  if (matchtag("coherence")) {
    parse_coherencematrix(matname);
    return;
  }
//   if (matchtag("exchange")) {
//     parse_generalmatrix(matname,false);
//     return;
//   }
  if (matchtag("general")) {
    parse_generalmatrix(matname);
    return;
  }
  
  setableoperator_spec* specp=create_setableoperator(get_curline());
  BlockedOperator* opexprp=new BlockedOperator();
//   if (specp->isconstant()) {    
//     *opexprp=masterobjp->make(specp->current());
//     delete specp; //!< can now delete definition
//     specp=NULL;
//   }
  operator_def* defp=new operator_def(specp,*opexprp);
  matrixmap[matname]=matrixspec(*defp);
  set_curline(NULL);
}
   
//! ensure all operator matrices are up-to-date
/** Not very clever for non-const operators - these are remade each time
    rather than being cached for one powder orienation to the next.
    Unlikely that this will be time limiting **/

void ensure_operator_matrices(bool constonly)
{
  if (!constonly)
    ensure_sigma0detect(); //!< sigma0/detect are special case
  const matrixmap_type::iterator end(matrixmap.end());
  matrixmap_type::iterator start(matrixmap.begin());
  while (start!=end) {
    matrixspec& mspec(start->second);
    if (!constonly || mspec.isconstant())
      mspec.update();
    ++start;
  }
}

bool matrixspec::isconstant() const
{
  switch (type) {
  case OPERATOR: {
    const operator_def& def(asoperator());
    return def.specp ? (def.specp)->isconstant() : false;
  }
  case EXCHANGE:
    return asexchange().isconstant();
  }
  return true;
}

void matrixspec::update()
{
  const bool isverb=((verbose & VER_GEN) && (verbose_level>1));

  switch (type) {
  case OPERATOR: {
    const operator_def& def(asoperator());
    if ( (&(def.op)==&sigma0) || (&(def.op)==&detect) || (def.specp==NULL))
      return; //!< ignore built-ins
    if ( !(def.specp->isconstant()) || !(def.op)) {
      const productoperator_spec spec(def.specp->current());
      if (isverb)
	std::cout << "Making operator: " << spec << '\n';
      def.op=masterobjp->make(spec); //!< if empty or non-constant, remake
    }    
  }
    break;
  case EXCHANGE: {
    ExchangeMatrix& mat(const_cast<ExchangeMatrix& >(asexchange()));
    if (!mat || (!(mat.isconstant())))
      mat.update();
  }
    break;
  case BOOL: {
    filter_def& def( const_cast<filter_def& >(asfilter()));
    if (!def) {
      if (isverb)
	std::cout << "Making filter matrix: " << def.spec << '\n';
      def.filterp=masterobjp->create_filter(def.spec);
    }
  }
    break;
  }
}
  
void post_channels()
{
  if (blockingnuclei.empty())
    error_abort("channels called before spin system");

  for (size_t i=nucids.size();i--;) {
    const size_t nuc=nucids(i);
    nuclei_spec* iter=::std::find(blockingnuclei.begin(),blockingnuclei.end(),nucids(i));
    if (iter==blockingnuclei.end()) {
      std::cerr << "RF channel " << nuctolabel(nuc) << " not present in spin system (or used twice)\n";
      error_abort(ERR_FAILED);
    }
    blockingnuclei.erase(iter);
  }
}

const basespin_system* get_spin_system()
{
  if (!sysp)
    error_abort("No spin system defined: missing spinsys block?");
  return sysp;
}

void make_pulseq_variables()
{
  add_systemvarmap(v_time);  
  if (spin_rate)
    add_systemvarmap(v_rotor_phase);
}

#endif

//double acqseq_synctime=0.0; //!< common synchronisation time for acqusition (0 if unset)

bool update_interactions=true;

size_t nchannels=0; //!< number of RF channels (0 before channels)
//double dphs=0.0; //!< receiver phase shift
double master_time=0.0; //!< current sequence time

setableoperator_spec* sigma0_specp=NULL; //!< initial density matrix specification (\c NULL if unset)
setableoperator_spec* detect_specp=NULL; //!< detect operator specification (\c NULL if unset)

//SystemVariable<double*> v_dphs("dphs",&dphs,rad_to_deg);
//SystemVariable<double*> v_synctime("acq_synctime",&acqseq_synctime,1e6);

LIST<size_t> nucids;

void add_channel(size_t nuc)
{
  nucids.push_back(nuc);
  nchannels=nucids.size();
}

void parse_channels()
{
  char* cptr;
  while ((cptr=parse_string(F_ALLOWMISSING)))
    add_channel(labeltonuc(cptr));

  post_channels();
}

Warning<> fullnotimp_warning("-full currently only supported for operator matrices.  Ignoring flag for: ",&NMRsim_once_warning);

void MasterObj::expandoperator(cmatrix& dest, const BlockedOperator& op) const
{
#ifndef NOPERIODIC
  if (!simple_opgenp)
    op.full(dest,*crystal_opgenp);
  else
#endif
    op.full(dest,*simple_opgenp);
}

void dump_matrix(const matrixspec& curm, const char* name, MasterObj* Objp, bool full)
{
  logfile_controller* logfilep(get_logfile());
  if ((curm.type==matrixspec::SPECIAL) && !Objp)
    error_abort("Hamiltonian is not defined at this point");
  if (full && (curm.type!=matrixspec::OPERATOR))
    fullnotimp_warning.raise(name);
  if (!logfilep) {
    std::cout << name << '\n';
    if (curm.type==matrixspec::SPECIAL)      
      Objp->printH(std::cout);
    else {
      if (full) {
	cmatrix tmp;
	Objp->expandoperator(tmp,curm.asoperator().op);
	std::cout << tmp;	
      }
      else
	std::cout << curm;
      if (verbose & VER_PROFILE)
	std::cout << curm.usage();
    }
  }
  else {
    char title[128];
    logfile_controller::maketitle(title,sizeof(title),name);
    if (curm.type==matrixspec::SPECIAL)
      Objp->logH(*logfilep,title);
    else
      curm.log(*logfilep,title,full);
  }
}

const flagsmap_type& putmatrix_flags()
{
  static flagsmap_type flags;
  if (flags.empty())
    flags["full"]=PM_FULL;
  return flags;
}

void parse_par_putmatrix()
{
  const char* name(parse_string());
  const matrixspec curm(findmap(matrixmap,name,"matrix"));
  const int flags=parse_flags(putmatrix_flags());
  dump_matrix(curm,name, NULL, flags & PM_FULL);
}

void make_1d_par_variables()
{
  make_data_variables();
  command_Factory_t& par_Factory(get_par_Factory());
  par_Factory["putmatrix"]=par_t(&parse_par_putmatrix,true);
  add_systemvarmap(v_proton_freq);
  add_systemvarmap(v_cputime);
}

context_t evaluation_state=CONTEXT_PRE; //!< pre evaluation

void initialise_simulation_environment()
{
  //adjust diagonalisation environment
  cmatrix_eigensystem_controller.tolerance = nochecks ? 0.0 : eigtolerance;
  cmatrix_eigensystem_controller.effectivezero = zerotolerance;
  int everbose = ((verbose & VER_GEN) && (verbose_level>1)) ? verbose_level : 0;
  if ((everbose==0) && abortonwarning) //!< ensures that matrices will be dumped
    everbose=1;
  cmatrix_eigensystem_controller.verbose = everbose;
  cmatrix_eigensystem_controller.throwexception = abortonwarning;
  evaluation_state=CONTEXT_PRE;
}

void MasterObj::initialise_simple_simulation()
{
  restart();
  ::initialise_simulation_environment();
  recreate_Hs(); //!< set up Hamiltonians
}

void parse_detect_operator()
{  
  detect_specp=parse_setableoperator();
}

void parse_start_operator()
{
  sigma0_specp=parse_setableoperator();
}

const productoperator_spec& setableoperator_spec::current() const
{ 
  return (size()==1)
    ? front()
    : (*this)(var_index % size());
}

productoperator_spec& setableoperator_spec::current()
{ 
  return (size()==1) ? front() : (*this)(var_index % size());
}

  //Don't support totalcoherence as this isn't very meaningful
ListList<int> parse_coherencelist()
{
  if (count_left()!=nchannels) {
    parser_printcontext() << "number of remaining arguments (" << count_left() << ") doesn't match number of RF channels (" << nchannels << ") - check directive syntax\n";
    error_abort();
  }
  ListList<int> cohers;
  for (size_t n=0;n<nchannels;n++)
    cohers.push_back(parse_intarray());
  if (verbose & VER_PARSE)
    std::cout << "Coherence list: " << cohers << '\n';
  return cohers;
}

BlockedFilter* MasterObj::create_filter(const ListList<int>& cohers) const
{ 
#ifndef NOPERIODIC
  if (!simple_opgenp)
    return new BlockedFilter(*crystal_opgenp,nucids,cohers);
  else
#endif
    return new BlockedFilter(*simple_opgenp,nucids,cohers);
}

void parse_coherencematrix(const char* matname)
{
  if (nchannels==0)
    error_abort("filters can only be applied to RF channels (no RF!)");
  
  filter_def* fdefp=new filter_def(parse_coherencelist());
  //  BlockedFilter* newf=masterobjp->create_filter(cohers);

  //  if (verbose & VER_GEN)
  //  std::cout << "Filter matrix " << matname << '\n' << (*newf) << '\n';
  matrixmap[matname]= matrixspec(*fdefp);
}

void parse_makefilter()
{
  parse_coherencematrix(parse_string());
}
  
void check_optional(const optional_t& request, const char* desc, bool val, const char* name)
{
  if (name==NULL)
    name=desc;  
  assert(get_optional_map().find(name)->second==&request); // if debugging check that name matches map
  if (verbose & VER_GEN)
    std::cout << "Using " << desc << " (" << name << "): " << (val ? "Yes\n" : "No\n");  
  if ((request==OPTIONAL_AUTO) || ((request==OPTIONAL_ON)==val))
    return;
  std::cerr << "Warning: " << ((request==OPTIONAL_OFF) ? "dis" : "en") << "able:" << name << " request was not fulfilled\n";
}

std::ostream& operator<< (std::ostream& ostr, const usage_t& usage)
{
  ostr << (usage.bytes/1024.0);
  if (usage.bytes)
    ostr << " K (" << usage.items << " items)";
  return ostr;
}

template<class T> void makeU(BlockedMatrix<complex>& U, double t1, double t2, const T& H, double maxdt =0.0)
{
  if (!H) {
    U.clear();
    return;
  }
  const int verb = (verbose & VER_GEN) ? verbose_level : 0;
  MetaPropagator propgen(H,maxdt,verb,global_prop_flags);
  set_partitioning(propgen);
  propgen(U,t1,t2);
}

void MasterObj::propagator(BlockedMatrix<complex>& U, double t1, double t2)
{
  switch (hamobjtype) {
  case M_SPIN:
    if (Hspinp.iscomplex())
      makeU(U,t1,t2,*(Hspinp.get_complex()),get_maxdt());
    else
      makeU(U,t1,t2,*(Hspinp.get_real()),get_maxdt());
    break;
  case M_STATICD:
    makeU(U,t1,t2,*Hstaticdp);
    break;
  case M_SPIND:
    makeU(U,t1,t2,*Hspindp);
    break;
  case M_STATIC:
    if (Hstaticp.iscomplex())
      makeU(U,t1,t2,*(Hstaticp.get_complex()));
    else
      makeU(U,t1,t2,*(Hstaticp.get_real()));
    break;
  default:
    throw Failed("Unknown hamobjtype");
  }
//   BlockedMatrix<complex> Utrans;
//   flush_transients(Utrans); //need to incorporate any unflushed transients
//   if (!!Utrans)
//     U*=Utrans;//U = U * Utrans (i.e. Utrans applied first)
}

bool have_spinsys() { return (sysp!=NULL); }

ExchangeMatrix::ExchangeMatrix(const char* namev, const BaseList<double>& rawvalsv, bool isconstv, bool isgeneralv) :
  name_(namev),
  sites_(0U),
  rawvals_(rawvalsv),
  isconst_(isconstv),
  isgeneral_(isgeneralv)
{
  update();
}

// ExchangeMatrix::ExchangeMatrix(const char* namev, const BaseList<double>& rawvalsv, const BaseList<double>& rawvalsiv, bool isconstv) :
//   name_(namev),
//   sites_(0U),
//   rawvalsr_(rawvalsv),
//   rawvalsi_(rawvalsiv),
//   isconst_(isconstv),
//   isgeneral_(true)
// {
//   update();
// }

Warning<> ExchangeMatrix::unphysical_warning("exchange matrix elements are expected to lie between 0 and 1: ",&NMRsim_once_warning);

void ExchangeMatrix::checkset(size_t r, size_t c, double v)
{
  char buf[256];
  if (v<0.0 || v>1.0) {
    snprintf(buf,sizeof(buf),"found %g when filling element %i,%i",v,r+1,c+1);
    unphysical_warning.raise(buf);
  }
  matrix_(r,c)=v;
}

//!< quick table-based integer sqrt
size_t intsqrt(size_t n)
{
  const size_t defmin=25U; //!< by default find roots up to 25
  static LIST<size_t> cache;
  if (cache.size()<=n) {
    size_t todo=(n<defmin) ? defmin : n;
    cache.create(todo+1,size_t(0));
    for (size_t i=0;;i++) {
      const size_t s=i*i;
      if (s>todo)
	break;
      cache(s)=i;
    }
  }
  return cache(n);
}

void ExchangeMatrix::update()
{
  size_t nels=rawvals_.size();
//   if (nels==0) { //!< special case of purely imaginary
//     nels=rawvalsi_.size();
//     if (nels==0) {
//       matrix_.clear();
//       return;
//     }
//     rawvalsr_.create(nels,0.0);
//   }
//   const bool isr=isreal();
//   if (!isr && (rawvalsi_.size()!=nels))
//     throw Mismatch("ExchangeMatrix::update");
 
  size_t n=0U;
  bool isr=true;
  if (isgeneral_) {
    n=intsqrt(nels);
    if (n==0) {
      if (((nels & 1)==0)) {
	n=intsqrt(nels/2);
	isr=false;
      }
      if (n==0) {
	parser_printcontext() << "number of supplied elements (" << nels << ") can't be number of elements in a square real or complex matrix\n";
	error_abort();
      }
    }
  }
  else {
    const size_t offdiagels=(nels & 1) ? 2*nels : nels;
    n=(1+intsqrt(4*offdiagels+1))/2;
    const size_t nhalf=n*(n-1)/2;
    if ((nels!=nhalf) && (nels!=2*nhalf)) {
      parser_printcontext() << "number of supplied matrix elements (" << nels << ") doesn't make sense as either the upper triangle or all off-diagonal elements of an exchange matrix\n";
      error_abort();
    }
//     if (!isr)
//       throw InternalError("exchange-only matrix must be purely real");  
  }
    
  matrix_.create(n,n);
  isreal_=isr;
  if (isgeneral_) {
    BaseList<complex> dest(matrix_.row());
    if (isr)
      dest=rawvals_;
    else {
      size_t i,j;
      for (i=j=0;j<nels;i++,j+=2)
	dest(i)=complex(rawvals_(j),rawvals_(j+1));
    }
    return;
  }
  const bool full=((nels & 1)==0);
  size_t r=0;
  size_t c=1;
  for (size_t i=0;i<nels;i++) {
    const double v=rawvals_(i);
    checkset(r,c,v);
    matrix_(r,c)=v;
    if (!full)
      matrix_(c,r)=v;
    if (++c==n) {
      r++;
      c= full ? 0U : r+1;
    }
    else {
      if (c==r)
	c++; //skip diagonal
    }
  }
  for (size_t c=n;c--;) {
    double diagv=1.0;
    for (r=n;r--;) {
      if (r!=c)
	diagv-=real(matrix_(r,c));
    }
    checkset(c,c,diagv);
  }
  if ((verbose & VER_GEN) && (verbose_level>1))
    std::cout << "Updated exchange matrix " << name_ << " from elements: " << rawvals_ << "\nto give matrix\n" << matrix_;
} 

void ExchangeMatrix::set(const BaseList<double>& vals, subsid_t)
{
  rawvals_=vals;
  update();
}

void ExchangeMatrix::printvariablename(std::ostream& ostr, subsid_t subsid) const
{
  ostr << name_;// << ((subsid==S_ARG1) ? "_real" : "_imag");
}
  
std::ostream& operator<< (std::ostream& ostr, const ExchangeMatrix& a)
{
  if (!a)
    return ostr << "[undefined in this context]\n";
  if (a.isreal())
    return ostr << real(a());
  else
    return ostr << a();
}

void ExchangeMatrix::printraw(std::ostream& ostr) const
{
  if (!(*this)) {
    ostr << "[undefined in this context]\n";
    return;
  }
  ostr << rawvals_;
  //if (!isreal())
  //  ostr << ' ' << rawvalsi_;
}
